#include "Walker.h"

#include "Solver.h"

#include <algorithm>

#include <cassert>

#define LOG(...)                                                               \
   do { } while (0)

#define LOG2(...)                               \
  do {                                          \
    printf(__VA_ARGS__);                        \
  } while (0)

namespace Glucose {

// These are in essence the CB values from Adrian Balint's thesis.  They
// denote the inverse 'cb' of the base 'b' of the (probability) weight
// 'b^-i' for picking a literal with the break value 'i' (first column is
// the 'size', second the 'CB' value).

static double cbvals[][2] = {
    {0.0, 2.00}, {3.0, 2.50}, {4.0, 2.85}, {5.0, 3.70},
    {6.0, 5.10}, {7.0, 7.40}, // Adrian has '5.4', but '7.4' looks better.
};

static const int ncbvals = sizeof cbvals / sizeof cbvals[0];

// We interpolate the CB values for uniform random SAT formula to the non
// integer situation of average clause size by piecewise linear functions.
//
//   y2 - y1
//   ------- * (x - x1) + y1
//   x2 - x1
//
// where 'x' is the average size of clauses and 'y' the CB value.

inline static double fitcbval(double size) {
  int i = 0;
  while (i + 2 < ncbvals && (cbvals[i][0] > size || cbvals[i + 1][0] < size))
    i++;
  const double x2 = cbvals[i + 1][0], x1 = cbvals[i][0];
  const double y2 = cbvals[i + 1][1], y1 = cbvals[i][1];
  const double dx = x2 - x1, dy = y2 - y1;
  assert(dx);
  const double res = dy * (size - x1) / dx + y1;
  assert(res > 0);
  return res;
}

CRef Walker::walk_pick_clause ()
{
  assert(!broken.empty());
  int64_t size = broken.size();
  if(size > INT_MAX) size = INT_MAX;
  int pos = random.pick_int(0, size - 1);

  return broken[pos];
}
Walker::Walker(Glucose::Solver * solver, double size, int64_t l, uint64_t seed)
  : solver (solver)
  , limit(l)
{
  random += seed;
  // This is the magic constant in ProbSAT (also called 'CB'), which we pick
  // according to the average size every second invocation and otherwise
  // just the default '2.0', which turns into the base '0.5'.
  //
  const bool use_size_based_cb = (solver->walk_count & 1);
  const double cb = use_size_based_cb ? fitcbval(size) : 2.0;
  assert(cb);
  const double base = 1 / cb; // scores are 'base^0,base^1,base^2,...

  double next = 1;
  for (epsilon = next; next; next = epsilon * base)
    table.push_back(epsilon = next);
}

  unsigned Solver::walk_break_value(Lit lit) {
    assert(value(lit) == l_True);
    unsigned res = 0;

    vec<Watcher> &ws = watches[~lit];
    auto end = ws.size();

    for (auto j = 1; j < end; ++j) {
      CRef cr = ws[j].cref;
      Lit blocker = ws[j].blocker;

      if(value(blocker) == l_True) continue;

      Clause &c = ca[cr];
      if(c.size() == 2) continue;

      assert(lit == c[0]);

      // Now try to find a second satisfied literal starting at 'literals[1]'
      // shifting all the traversed literals to right by one position in order
      // to move such a second satisfying literal to 'literals[1]'.  This move
      // to front strategy improves the chances to find the second satisfying
      // literal earlier in subsequent break-count computations.
      //

      auto begin = 1;
      const auto end = c.size();
      auto i = begin;
      Lit prev = mkLit(INT_MAX);

      while (i < end) {
        const Lit other = c[i];
        c[i ++] = prev;
        prev = other;
        if (value(other) == l_False) continue;

        // Found 'other' as second satisfying literal.
        ws[j].blocker = other; // Update 'blit'
        c[begin] = other;        // and move to front.
        break;
      }

      if (i != end)
        continue; // Double satisfied!

      // Otherwise restore literals (undo shift to the right).
      //
      while(i != begin) {
        const Lit other = c[--i];
        c[i] = prev;
        prev = other;
      }

      res++; // Literal 'lit' single satisfies clause 'c'.
    }

    return res;
  }

Lit Solver::walk_pick_lit(Walker & walker, const Clause &c) {
  assert(walker.int_score.empty());
  LOG("\nc walk_pick_lit in clause:");
  printClause(c);
  LOG("\n");
  double sum = 0;
  int64_t propagations = 0;
  int64_t size = c.size();
  int begin = 0;
  for(auto i = begin; i < size; ++i) {
    const Lit lit = c[i];
    if(level(var(lit)) == 0)
      continue;

    ticks++; ++propagations;
    unsigned tmp = walk_break_value(~lit);
    double score = walker.score(tmp);
    walker.int_score.push_back(score);
    sum += score;
  }
  assert(walker.int_score.size() != 0);
  walker.propagations += propagations;

  assert(walker.int_score.size() <= c.size());

  const double lim = sum + walker.random.generate_double();

  auto i = 0;
  auto j = walker.int_score.begin();
  const auto end = c.size();
  Lit res = mkLit(INT_MAX);
  for(;;) {
    assert(i < end);
    res = c[i++];
    if(level(var(res)) > 0) break;
    LOG("skipping var %d\n", var(res));
  }

  sum = *j++;
  while (sum <= lim && i < end) {
    if (level(var(c[i])) == 0) {
      LOG("skipping var %d without score\n", var(res));
      ++i;
      continue;
    }
    res = c[i++];
    LOG("considering var %d at position %d\n", var(res), i - 1);
    assert(j < walker.int_score.end());
    sum += *j++;
  }

  walker.int_score.clear();
  return res;
}


// The scores are tabulated for faster computation (to avoid 'pow').

inline double Walker::score(unsigned i) {
  const double res = (i < table.size() ? table[i] : epsilon);
  return res;
}

void Solver::walk_move_satisfied_lit (Walker &walker, CRef cr)
{
  printClause(ca[cr]);
  LOG("\n");
  bool satisfiable = false;
  int satisfied = 0;
  bool failed = false;

  Clause &c = ca[cr];
  int size = c.size();

  for(int i = 0; satisfied < 1 && i < size; ++i) {
    const Lit lit = c[i];
    assert(value(lit) == l_True || value(lit) == l_False);
    if (value(lit) == l_True) {
      auto lit2 = c[satisfied];
      c[satisfied] = c[i];
      c[i] = lit2;
      LOG("c swap %d and %d\n", satisfied, i);
      printClause(ca[cr]);
      satisfied++;
    } else if(!satisfiable && level(var(lit)) > 0) {
      satisfiable = true;
    }
  }

  if(!satisfied && !satisfiable) {
    failed = true;
  }

  if (satisfied) {
    printClause(ca[cr]);
    LOG("\n");
    fflush(stdout);
    assert(value(c[0]) == l_True);
    watches[~c[0]].push(Watcher(cr, c[1]));
    LOG ("c pushing clause co WL ");
    printClause(c);
    LOG("(there are %d other clauses)\n", watches[~c[0]].size() - 1);
  } else {
    assert(value(c[0]) == l_False);
    walker.broken.push_back(cr);
    LOG("c pushing broken clause ");
    printClause(c);
    LOG("\n");
  }
}

void Solver::walk_save_minimum(Walker & walker) {
  size_t broken = walker.broken.size();
  if(broken >= walk_minimum) return;
  walk_minimum = broken;
  LOG("c walk: now only %ld broken", walk_minimum);
  polarity.memcopyTo(min_polarity);
}

bool Solver::walk_propagate(CRef cr, Lit lit)
{
  Lit prev = mkLit(-1);
  ++ticks;
  Clause &c = ca[cr];

  LOG("c propagating clause ");
  printClause(c);
  LOG("\n");
  // find lit in c
  for(int i = 0; i < c.size(); ++i) {
    const Lit other = c[i];
    c[i] = prev;
    prev = other;
    if(other == lit) break;
    assert(value(other) == l_False);
  }

  if(prev == lit) {
    c[0] = lit;
    watches[~c[0]].push(Watcher(cr, c[1]));
    return true;

  } else { // not satisfied, undo the shifting
    for(int i = c.size() - 1; i >= 0; --i) {
      const Lit other = c[i];
      c[i] = prev;
      prev = other;
    }
    assert(c[0] != mkLit(-1));
    return false;
  }

}


void Solver::walk_flip_lit(Walker &walker, Lit lit)
{

    assert(value(lit) == l_False);
    assigns[var(lit)] = lbool(!sign(lit));
    assert(value(lit) == l_True);

    LOG("c updating %d broken clauses\n", walker.broken.size());
    {
      int j = 0;
      for (auto i = 0; i < walker.broken.size(); ++i, ++j) {
        walker.broken[j] = walker.broken[i];
        const bool moved_to_WL = walk_propagate(walker.broken[i], lit);
        if(moved_to_WL){
          LOG("c deleted as ");
          printClause(ca[walker.broken[j]]);
          LOG("\n");
          --j;
        }
      }

      walker.broken.resize(j);
    }

    LOG("c updating new unsatisfied clauses\n");
    // Finally add all new unsatisfied (broken) clauses.
    {

      vec<Watcher> &ws = watches[lit];
      Watcher *i, *j, *end;
      Lit prev = mkLit(-1);

      for (i = (Watcher *)ws, end = i + ws.size(); i != end; ++i) {
        Lit replacement;
        bool replacement_found = false;

        CRef cr = i->cref;
        Clause &c = ca[cr];

        LOG("c updating clause ");
        printClause(c);
        LOG("\n");
        const int size = c.size();
        const Lit lit2 = c[0];
        assert(lit2 == ~lit);
        for (int i = 1; i < size; ++i) {
          const Lit other = c[i];
          c[i] = prev;
          prev = other;
          const auto v = value(other);
          assert(v != l_Undef);
          if(v == l_False)
            continue;
          replacement = other;
          replacement_found = true;
          break;
        }

        if(replacement_found) {
          c[1] = ~lit;
          c[0] = replacement;
          assert(c[1] != c[0]);
          LOG ("c pushing clause ");
          printClause(c);
          LOG("(there are %d other clauses)\n", watches[~c[0]].size());
          watches[~c[0]].push(Watcher(cr, c[1]));
        } else {
          for (int i = size - 1; i > 0; --i) {
            const Lit other = c[i];
            c[i] = prev;
            prev = other;
          }
          assert(c[0] == ~lit);
          walker.broken.push_back(cr);
          walker.broken_count++;
          LOG("c pushing broken clause ");
          printClause(c);
          LOG("\n");
        }
      }
      ws.shrink(ws.size());
    }
}

int Solver::walk_round(int64_t limit, bool prev) {
  cancelUntil(0);
  simplify();
  newDecisionLevel();
  int rounds = 0;

  for (auto i = 0; i < clauses.size(); ++i) {
    printClause(clauses[i]);
  }

  if (walkredundant) {
    for (auto i = 0; i < learnts.size(); ++i) {
      printClause(learnts[i]);
    }
  }

  LOG("c empty watches\n");
  for (int v = 0; v < nVars(); v++) {
    watches[mkLit(v, true)].shrink(watches[mkLit(v, true)].size());
    watches[mkLit(v, false)].shrink(watches [mkLit(v, false)].size());
    watchesBin[mkLit(v, true)].shrink(watchesBin[mkLit(v, true)].size());
    watchesBin[mkLit(v, false)].shrink(watchesBin[mkLit(v, false)].size());
  }

  learnts_literals = 0;
  clauses_literals = 0;
  double size = 0;
  int64_t n = 0;

  for (auto i = 0; i < clauses.size(); ++i) {
    Clause &c = ca[clauses[i]];
    size += c.size();
    ++n;
  }

  if(walkredundant) {
    for (auto i = 0; i < learnts.size(); ++i) {
      Clause &c = ca[learnts[i]];
      if (c.lbd() > 6) continue;
        size += c.size();
      ++n;
    }
  }

  for (int v = 0; v < nVars(); v++)
    if(value(v) == l_Undef) {
      const bool phase = prev ? prev_polarity[v] : decide_phase(v, true);
      assigns[v] = phase ? l_True : l_False;
      vardata[v].level = 1;
    } else {
      assert(vardata[v].level == 0);
      LOG("var %d is set\n", v);
    }


  double average_size = (double)size / (double)n;

  Walker walker(this, average_size, limit, walk_count);

  LOG("c split clauses\n");
  for (auto i = 0; i < clauses.size(); ++i) {
    walk_move_satisfied_lit(walker, clauses[i]);
  }

  if (walkredundant) {
    for (auto i = 0; i < learnts.size(); ++i) {
      if (ca[learnts[i]].lbd() > 6)
        continue;
      walk_move_satisfied_lit(walker, learnts[i]);
    }
  }


  int64_t old_minimum = walk_minimum;

  int64_t broken = walker.broken.size();
  int64_t flips = 0,
    minimum = broken;

  LOG2("c Starting walker has %ld broken clauses (%d rounds, best: %ld)\n", minimum,
       rounds, walk_minimum);

  LOG("c start walking\n");
  while(ticks < limit &&
        !walker.broken.empty()) {
    ++flips;
    ++rounds;
    CRef c = walker.walk_pick_clause();
    LOG("c picked clause ");
    printClause(ca[c]);
    LOG("\n");
    for (int i = 0; i < ca[c].size(); ++i) {
      const Lit other = ca[c][i];
      assert(value(other) == l_False);
    }

    const Lit lit = walk_pick_lit(walker, ca[c]);
    assert(value(lit) == l_False);
    assert (level(var(lit)) > 0);
    LOG ("c flipping lit %d of value %d (%ld broken)\n", toInt(lit), value(lit) == l_True ? 1 : 0,
                      walker.broken.size());
    walk_flip_lit(walker, lit);
    walker.broken_count = walker.broken.size();
    if(walker.broken.size() >= minimum) continue;
    minimum = walker.broken.size();
    polarity.memcopyTo(saved_polarity);
    walk_save_minimum(walker);
  }

  if(minimum < old_minimum)
    LOG("found better minimum %d", minimum);

  polarity.memcopyTo(prev_polarity);

  // if (walk_minimum == 0) {
  //   LOG("c \n");
  //   for (auto i = 0; i < clauses.size(); ++i) {
  //     const Lit lit = ca[clauses[i]][0];
  //     if (!satisfied(ca[clauses[i]])) {
  //       printClause(ca[clauses[i]]);
  //       fflush(stdout);
  //       assert(false);
  //     };
  //   }

  //   for (auto i = 0; i < learnts.size(); ++i) {
  //     const Lit lit = ca[learnts[i]][0];
  //     assert(value(lit) == l_True);
  //   }
  //   LOG("c found model:\n");
  //   for (int v = 0; v < nVars(); v++) {
  //     if (polarity[toInt(v)])
  //       LOG("%d 0\n", v + 1);
  //     else
  //       LOG("-%d 0\n", v + 1);
  //   }
  //   abort();
  // }

    int res = 0;
    if (minimum == 0) {
      LOG("c walking found model");
      res = 10;
    }
    for (int v = 0; v < nVars(); v++)
      if (level(v) > 0) {
        assigns[v] = l_Undef;
      }
    for (int v = 0; v < nVars(); v++)
      if (value(v) != l_Undef) {
        LOG("var %d is set to %d\n", v + 1, (value(v) == l_True ? 1 : -1));
      }

    for (int v = 0; v < nVars(); v++) {
      watches[mkLit(v, true)].shrink(watches[mkLit(v, true)].size());
      watches[mkLit(v, false)].shrink(watches[mkLit(v, false)].size());
      watchesBin[mkLit(v, true)].shrink(watchesBin[mkLit(v, true)].size());
      watchesBin[mkLit(v, false)].shrink(watchesBin[mkLit(v, false)].size());
    }


  cancelUntilNoSaving(0);

  for (auto i = 0; i < clauses.size(); ++i) {
    attachClause(clauses[i]);
    printClause(clauses[i]);
  }

  for (auto i = 0; i < learnts.size(); ++i) {
    attachClause(learnts[i]);
    printClause(learnts[i]);
  }


  auto m = 0;

  for (int v = 0; v < nVars(); v++) {
    m += watches[mkLit(v, true)].size() +
    watches[mkLit(v, false)].size() +
    watchesBin[mkLit(v, true)].size() +
    watchesBin[mkLit(v, false)].size();
  }
  LOG2("c Walker has %ld broken clauses (%d rounds, best: %ld)\n",
       minimum, rounds, walk_minimum);

  return res;
}

void Solver::walk() {
  profiler[int_of_mode(mode)].end();
  MODE old_mode = mode;
  mode = MODE::WALK;
  profiler[int_of_mode(mode)].start();
  const int budget = 10 * (learnts.size() + clauses.size()); // 1'500'000;
  int64_t limit = ticks + budget;
  (void)walk_round(limit, false);
  ++walk_count;
  profiler[int_of_mode(mode)].end();
  mode = old_mode;
  profiler[int_of_mode(mode)].start();
}
} // namespace Glucose
